module lrg_pk
 use CAMB
 use CMB_Cls
 implicit none

 integer, parameter :: lrg_maxnpoints=8192
 integer, parameter :: Nsample = 9
 real, parameter :: PI_S = 3.1415926535897932384626434E0

!-------- Define a structure to hold power spectrum data
 Type PkData
    integer  :: npoints
    real :: minkh, dlnkh, nn, sigma8
    real, dimension(1:Nsample):: zval 
    type(CAMBparams) :: P
    real, dimension(lrg_maxnpoints,2) :: pkarr    ! pkarr(i,1) is the k, pkarr(i,2) is the Pk
    real, dimension(lrg_maxnpoints,2,1:Nsample) :: pkarr_Nsample
    real, dimension(lrg_maxnpoints,3,1:Nsample) :: spkarr
 end Type PkData

  
!-------- Spline data for growth factor
  integer, parameter :: nzspline = 100
!  real, parameter :: dz_spline = 0.01
  real :: growlz(nzspline), growl(nzspline), growl2(nzspline)


 public
 private :: spline, splint, lrg_maxnpoints, wint

 contains

   subroutine LRG_ComputePkData(Pk, eval)
!-------- Allocate a structure given a minimum kh, npoints, and
! logarithmic spacing...., then call CAMB if necessary, and compute the
! power spectrum.
! The power spectrum is computed at the present time (to allow for easy 
! normalizations)
     implicit none

     type(PkData) :: Pk
     integer :: i, eval, np
     real:: kmax
     real, dimension (1:Nsample):: norm
     type(CAMBparams) :: P
     real, dimension(lrg_maxnpoints,3) :: spkarr_tmp
     character (len=20) :: time1,date1
     real :: time

!-------- Check to see if npoints is exceeded
     if (Pk%npoints .GT. lrg_maxnpoints) stop 'lrg_maxnpoints exceeded'
     np=Pk%npoints


!-------- Copy data
     P = Pk%P
     P%WantTransfer= .true.
     P%WantCls = .false.
     kmax = Pk%minkh*exp((Pk%npoints -1)*Pk%dlnkh)
     P%Transfer%high_precision=.false.  
     P%Transfer%kmax=2*kmax
     P%Transfer%k_per_logint=5
     P%Transfer%num_redshifts=1
     P%Transfer%redshifts(1)=0
     P%OutputNormalization = outNone
     P%InitPower%nn     = 1 !number of initial power spectra
     P%InitPower%ant(1) = 0 !Not used here
     P%InitPower%rat(1) = 1 !ditto
     P%InitPower%an(1) = Pk%nn

!-------- Set up the kh array
     do i=1, Pk%npoints
        Pk%pkarr(i,1) = Pk%minkh*exp((i-1_dl)*Pk%dlnkh)
     end do
     
!-------- Call CAMB if necessary, otherwise use cached value
     if (eval .ne. 0) then 
        call CAMB_GetResults(P)
     else 
!-------- The following updates the arrays
        call CAMBParams_Set(P)
        call InitializePowers(P%InitPower, P%curv)
     end if

!-------- Compute sigma8 and find the normalizing factor
! This is mildly inefficient, but the losses are small compared to the 
! cost of computing the transfer functions, so we will ignore it.
     call Transfer_Get_sigma8(MT, 8.0d0)
      norm(1) = (Pk%sigma8/MT%sigma_8(1,1))
!-------- Now compute the power spectrum
     call Transfer_GetMatterPower(MT,Pk%pkarr(1:np,2), 1, 1, Pk%minkh, &
          Pk%dlnkh, Pk%npoints)
       Pk%pkarr_Nsample(1:np,1,1) = Pk%pkarr(1:np,1)
       Pk%pkarr_Nsample(1:np,2,1) = Pk%pkarr(1:np,2)
       Pk%pkarr_Nsample(1:np,2,1) = Pk%pkarr(1:np,2)*(norm(1)**2)*Pk%pkarr(1:np,1)**3 / (2_dl*PI_S*PI_S)
!-------- Now spline the logarithm of the power spectrum
     Pk%spkarr(1:np,1:2,1) = log(Pk%pkarr_Nsample(1:np,1:2,1))
!-------- we need to do a loop here in order to spline all the spkarr for each sample!
     do i=1, 1
       spkarr_tmp(1:np,1:2) = Pk%spkarr(1:np,1:2,i)

       call spline(spkarr_tmp(1:np,1), spkarr_tmp(1:np,2), Pk%npoints, &
          1.e30,1.e30,spkarr_tmp(1:np,3))
       
       Pk%spkarr(1:np,3,i) = spkarr_tmp(1:np,3)

     end do
    
   end subroutine LRG_ComputePkData

   real function LRG_PkSpline(kh, Pk, iSample)

     implicit none
     type(PkData) :: Pk
     real :: kh,  lnkh
     real:: tmp
     integer :: np,i,iSample
     real, dimension(lrg_maxnpoints,3) :: spkarr_tmp



     np=Pk%npoints
     lnkh = log(kh)

     spkarr_tmp(1:np,1:3) = Pk%spkarr(1:np,1:3,iSample)
     call splint(spkarr_tmp(1:np,1),spkarr_tmp(1:np,2),spkarr_tmp(1:np,3),Pk%npoints,lnkh,tmp)
     LRG_PkSpline = exp(tmp)

   end function LRG_PkSpline


   subroutine LRG_ComputePkNlin(Pklin, PkNlin,iSample)
     
     implicit none
     type(PkData), intent(in) :: Pklin
     type(PkData) :: PkNlin
     real :: xlogr1, xlogr2, diff, rmid, sig, d1, d2, rncur, rknl, rneff
     real :: pnl, ph, pq, rk, plin
     integer :: i, np, iSample
     real, dimension(lrg_maxnpoints,3) :: spkarr_tmp
  

!-------- First just copy the data across
      PkNlin%npoints = Pklin%npoints
      PkNlin%minkh = Pklin%minkh
      PkNlin%dlnkh = Pklin%dlnkh
      PkNlin%nn = Pklin%nn
      PkNlin%sigma8 = Pklin%sigma8
      PkNlin%zval(iSample) = Pklin%zval(iSample)
      PkNlin%P = Pklin%P
      PkNlin%pkarr = Pklin%pkarr
      PkNlin%pkarr_Nsample(:,:,iSample) = Pklin%pkarr_Nsample(:,:,iSample)
      PkNlin%spkarr(:,:,iSample) = Pklin%spkarr(:,:,iSample)


     
     do i = 1, Pklin%npoints
       plin = Pklin%pkarr_Nsample(i,2,iSample)
        PkNlin%pkarr_Nsample(i,2,iSample) = plin  ! we are not using halofit!! just using the linear powerspectrum!!
     end do
!-------- Now spline the logarithm of the power spectrum
     np = PkNlin%npoints
     PkNlin%spkarr(1:np,1:2,iSample) = log(PkNlin%pkarr_Nsample(1:np,1:2,iSample))
    
     spkarr_tmp(1:np,1:2) = PkNlin%spkarr(1:np,1:2,iSample)

     call spline(spkarr_tmp(1:np,1),spkarr_tmp(1:np,2), PkNlin%npoints, &
          1.e30,1.e30,spkarr_tmp(1:np,3))
     PkNlin%spkarr(1:np,3,iSample) = spkarr_tmp(1:np,3)

   end subroutine LRG_ComputePkNlin


   real function lrg_growl(CMB,k,z)  !JD added k
!-------- Compute the growth function
     implicit none

     type(CMBParams) :: CMB
     real(dl) :: a0, lnamin, lna0, fac, tmp, ga, ua
     real(dl) :: del_lna, lna , lnastart, lnaend,lna1, lna2,lna3  
     real :: z, k !JD added k
     integer :: lnabin,i_lna
     real(dl) :: dua1_dlna, dga1_dlna, dua2_dlna, dga2_dlna
     real(dl) :: dua3_dlna, dga3_dlna, dua4_dlna, dga4_dlna
     real(dl) :: ua1, ga1, ua2, ga2, ua3, ga3,amin
    

     lnabin = 100 
     a0 = 1.0/(1.0 + z)
     lna0 = log(a0)
     amin = 1.0_dl/10000_dl
     lnamin = log(amin)
     del_lna = (lna0-lnamin)/lnabin

!initial conditions
     ga = amin  !Ga = a at matter dominated universe
     ua = 1.0_dl            !ua = 1 at matter dominated universe
! we need to integrate a couple differential equation here

     do i_lna=0,(lnabin-1)
 ! doing rk4th

      lna = lnamin+(i_lna*del_lna)

      
      dua1_dlna = func_du_dlna(CMB,k,lna,del_lna,ua,ga)  !JD added k for TGR scale dependence
      dga1_dlna = func_dg_dlna(lna,del_lna,ua,ga)
    
      ua1 =  ua+(dua1_dlna*del_lna/2.0)
      ga1 =  ga+(dga1_dlna*del_lna/2.0)
      lna1 = lnamin+(i_lna*del_lna+del_lna/2.0)

      dua2_dlna = func_du_dlna(CMB,k,lna1,del_lna/2.0,ua1,ga1) !JD added k for TGR scale dependence 
      dga2_dlna = func_dg_dlna(lna1,del_lna/2.0,ua1,ga1)

      ua2 =  ua+(dua2_dlna*del_lna/2.0)
      ga2 =  ga+(dga2_dlna*del_lna/2.0)
      lna2 = lnamin+(i_lna*del_lna+del_lna/2.0)

      dua3_dlna = func_du_dlna(CMB,k,lna2,del_lna/2.0,ua2,ga2)  !JD added k for TGR scale dependence
      dga3_dlna = func_dg_dlna(lna2,del_lna/2.0,ua2,ga2)
      
      ua3 =  ua+(dua3_dlna*del_lna)
      ga3 =  ga+(dga3_dlna*del_lna)
      lna3 = lnamin+(i_lna*del_lna+del_lna)
      
      dua4_dlna = func_du_dlna(CMB,k,lna3,del_lna,ua3,ga3)  !JD added k for TGR scale dependence
      dga4_dlna = func_dg_dlna(lna3,del_lna,ua3,ga3)     
      
      ga =  ga + (del_lna/6.0)*( dga1_dlna + 2.0*dga2_dlna + 2.0*dga3_dlna +  dga4_dlna)
      ua =  ua + (del_lna/6.0)*( dua1_dlna + 2.0*dua2_dlna + 2.0*dua3_dlna +  dua4_dlna)

     end do
     lrg_growl= ga

     contains


       real function lrg_hubble(CMB,a)
!-------- Just assumes matter, and Lambda
         implicit none
         real(dl) :: a, omfac, olfac, okfac, Ea
         type (CMBParams) CMB

         omfac = (CMB%omb + CMB%omc  + CMB%omnu )/(a**3) !matter
         olfac = CMB%omv/(a**3*(1+CMB%w))   !de, w allowed
         okfac = (1.0_dl - (CMb%omb+CMB%omc+CMB%omv))/(a**2) !curvature
         
         Ea = omfac+okfac+olfac

         lrg_hubble = sqrt(Ea)
       end function lrg_hubble


       real(dl) function lrg_hubble2(CMB,a)
!-------- Just matter and dark energy
         implicit none
         real(dl) :: a, omfac, olfac, okfac, Ea
         type (CMBParams) CMB

         omfac = (CMB%omb + CMB%omc + CMB%omnu)/(a**3.0_dl) !matter-like (baryons, cold dark matter, neutrinos
         olfac = CMB%omv/(a**(3.0_dl*(1.0_dl+CMB%w)))   !de, w allowed
         okfac = (1.0_dl - (CMB%omb+CMB%omc+CMB%omv + CMB%omnu))/(a**2.0_dl) !curvature
         
         Ea = omfac+okfac+olfac

         lrg_hubble2 = Ea

       end function lrg_hubble2
       
       real(dl) function func_du_dlna(CMB,k,lna,del_lna,ua,ga)  !JD added k
         implicit none
         real(dl) lna, num, den, a, aold, dlnEa_dlna, del_lna, ua, ga, Ea  
         real(dl) oldEa
         real(dl) :: geff
         real k  !JD added k
         external :: geff
         type(CMBParams) CMB

         a = exp(lna)
         aold = exp((lna-del_lna))
         Ea = lrg_hubble2(CMB,a)
         oldEa = lrg_hubble2(CMB,aold)
         dlnEa_dlna = (log(Ea)-log(oldEa))/del_lna
         func_du_dlna = -(3.0_dl+(0.5_dl*dlnEa_dlna))*ua + &
         (3.0_dl*geff(k,a)*(CMB%omb+CMB%omc)*ga/(2.0_dl*(a**4.0_dl)*Ea))
!         (3.0_dl*(CMB%omb+CMB%omc)*ga/(2.0_dl*(a**4.0_dl)*Ea))  !JD multiplied by geff(k,a) above for TGR
       end function func_du_dlna

       real(dl) function func_dg_dlna(lna,del_lna,ua,ga)
         implicit none
         real(dl) lna, num, den, a, del_lna, ua, ga

         a = exp(lna)
         func_dg_dlna = a*ua
     
       end function func_dg_dlna



     end function lrg_growl


!-------- Halofit code from R.E. Smith, modified for F90

! halo model nonlinear fitting formula as described in 
! Appendix C of Smith et al. (2002)

      subroutine halofit(rk,rn,rncur,rknl,plin,pnl,pq,ph, P)

        implicit none
        type(CAMBparams) :: P
        
        real(dl):: gam,a,amod,b,c,xmu,xnu,alpha,beta,f1,f2,f3,f4
        real ::  rk,rn,plin,pnl,pq,ph, rknl,y,rncur,p_cdm
        real(dl) :: f1a,f2a,f3a,f4a,f1b,f2b,f3b,f4b,frac, om_m, om_v
        

        gam=0.86485+0.2989*rn+0.1631*rncur
        a=1.4861+1.83693*rn+1.67618*rn*rn+0.7940*rn*rn*rn+ &
             0.1670756*rn*rn*rn*rn-0.620695*rncur
        a=10**a      
        b=10**(0.9463+0.9466*rn+0.3084*rn*rn-0.940*rncur)
        c=10**(-0.2807+0.6669*rn+0.3214*rn*rn-0.0793*rncur)
        xmu=10**(-3.54419+0.19086*rn)
        xnu=10**(0.95897+1.2857*rn)
        alpha=1.38848+0.3701*rn-0.1452*rn*rn
        beta=0.8291+0.9854*rn+0.3400*rn**2
        om_m = P%omegab + P%omegac
        om_v = P%omegav

        if(abs(1-om_m).gt.0.01) then ! omega evolution 
           f1a=om_m**(-0.0732)
           f2a=om_m**(-0.1423)
           f3a=om_m**(0.0725)
           f1b=om_m**(-0.0307)
           f2b=om_m**(-0.0585)
           f3b=om_m**(0.0743)       
           frac=om_v/(1.-om_m) 
           f1=frac*f1b + (1-frac)*f1a
           f2=frac*f2b + (1-frac)*f2a
           f3=frac*f3b + (1-frac)*f3a
        else         
           f1=1.0
           f2=1.
           f3=1.
        endif

        y=(rk/rknl)
        
        ph=a*y**(f1*3)/(1+b*y**(f2)+(f3*c*y)**(3-gam))
        ph=ph/(1+xmu*y**(-1)+xnu*y**(-2))
        pq=plin*(1+plin)**beta/(1+plin*alpha)*exp(-y/4.0-y**2/8.0)

        pnl=pq+ph

      end subroutine halofit



!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

! The subroutine wint, finds the effective spectral quantities
! rknl, rneff & rncur. This it does by calculating the radius of 
! the Gaussian filter at which the variance is unity = rknl.
! rneff is defined as the first derivative of the variance, calculated 
! at the nonlinear wavenumber and similarly the rncur is the second
! derivative at the nonlinear wavenumber. 

      subroutine wint(Pk,r,sig,d1,d2,iSample)
      
      implicit none
      type(PkData) :: Pk
      real(dl) :: sum1,sum2,sum3,t,y,x,w1,w2,w3
      real :: r, sig, d1,d2, rk
      integer i,nint,iSample


      
      nint=3000
      sum1=0.d0
      sum2=0.d0
      sum3=0.d0
      do i=1,nint
         t=(float(i)-0.5)/float(nint)
         y=-1.d0+1.d0/t
         rk=y
         d2=LRG_PkSpline(rk,Pk,iSample)
         x=y*r
         w1=exp(-x*x)
         w2=2*x*x*w1
         w3=4*x*x*(1-x*x)*w1
         sum1=sum1+w1*d2/y/t/t
         sum2=sum2+w2*d2/y/t/t
         sum3=sum3+w3*d2/y/t/t
      enddo
      sum1=sum1/float(nint)
      sum2=sum2/float(nint)
      sum3=sum3/float(nint)
      sig=sqrt(sum1)
      d1=-sum2/sum1
      d2=-sum2*sum2/sum1/sum1 - sum3/sum1
      
    end subroutine wint
      
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%






!--------  NUMERICAL RECIPES ROUTINES ARE BELOW

     SUBROUTINE spline(x,y,n,yp1,ypn,y2)
       INTEGER n,NMAX
       real yp1,ypn
       real x(n), y(n), y2(n)
       PARAMETER (NMAX=10000)	! Increased from 500 by Max
       INTEGER i,k
       real p,qn,sig,un,u(NMAX)
       if (N.gt.NMAX) pause 'SPLINE NMAX DEATH ERROR' ! Added by Max
       if (x(1).gt.x(n)) pause 'SPLINE WARNING: x NOT INCREASING' ! Added by Max
       if (yp1.gt..99e30) then
          y2(1)=0.
          u(1)=0.
       else
          y2(1)=-0.5
          u(1)=(3./(x(2)-x(1)))*((y(2)-y(1))/(x(2)-x(1))-yp1)
       endif
       do i=2,n-1
          sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
          p=sig*y2(i-1)+2.
          y2(i)=(sig-1.)/p
          u(i)=(6.*((y(i+1)-y(i))/(x(i+ &
               1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig &
               *u(i-1))/p
       end do
       if (ypn.gt..99e30) then
          qn=0.
          un=0.
       else
          qn=0.5
          un=(3./(x(n)-x(n-1)))*(ypn-(y(n)-y(n-1))/(x(n)-x(n-1)))
       endif
       y2(n)=(un-qn*u(n-1))/(qn*y2(n-1)+1.)
       do k=n-1,1,-1
          y2(k)=y2(k)*y2(k+1)+u(k)
       end do

     END SUBROUTINE spline


     SUBROUTINE splint(xa,ya,y2a,n,x,y)
       INTEGER n
       real x,y
       real xa(n),y2a(n),ya(n)
       INTEGER k,khi,klo
       real a,b,h

!-------- If out of bounds, then linearly extrapolate
! else use the cubic spline
      if (x .LT. xa(1)) then
         y = ya(1) + (ya(1)-ya(2))*(x-xa(1))/(xa(1)-xa(2)) 
      else if (x .GT. xa(n)) then
         y = ya(n) + (ya(n)-ya(n-1))*(x-xa(n))/(xa(n)-xa(n-1))
      else
         klo=1
         khi=n
         do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(xa(k).gt.x)then
               khi=k
            else
               klo=k
            endif
         end do
         h=xa(khi)-xa(klo)
         if (h.eq.0.) pause 'bad xa input in splint'
         a=(xa(khi)-x)/h
         b=(x-xa(klo))/h
         y=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))*(h** &
              2)/6.
      end if

     END SUBROUTINE splint



end module lrg_pk
